import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.apache.spark.api.java.JavaRDD;
import org.apache.spark.api.java.JavaSparkContext;
import org.apache.spark.api.java.function.Function;
import org.apache.spark.storage.StorageLevel;

public class Main {
	
	private static JavaRDD<Point> mData;
	
	// Data sample size that will be chosen from row data to process
	private static int DATA_SAMPLE_SIZE;
	// Number of medoids to process
	private static int NUM_OF_MEDOIDS;
	// Number of experiences to run
	private static int NUM_OF_EXPERIMENTS;
		
	/**
	 * @param args
	 */
	public static void main(String[] args) {
		double start = System.currentTimeMillis();
		String testDataPath = "data.csv";
		DATA_SAMPLE_SIZE = 40 + 6 * 2;
		NUM_OF_MEDOIDS = 5;
		NUM_OF_EXPERIMENTS = 10;
		int procNum = 1;
		
		if(args != null && args.length > 0 && args[0] != null) DATA_SAMPLE_SIZE = Integer.parseInt(args[0]);
		if(args != null && args.length > 1 && args[1] != null) NUM_OF_MEDOIDS = Integer.parseInt(args[1]);
		if(args != null && args.length > 2 && args[2] != null) NUM_OF_EXPERIMENTS = Integer.parseInt(args[2]);
		if(args != null && args.length > 3 && args[3] != null) procNum = Integer.parseInt(args[3]);
		if(args != null && args.length > 4 && args[4] != null) testDataPath = args[4];
		
		DateFormat dateFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss");
		Date date = new Date();
		JavaSparkContext sc = new JavaSparkContext("spark://laada-spark-master:7077", "Spark Clara App", "/root/spark-0.8.0-incubating", "SparkClara.jar");
		JavaRDD<String> data = sc.textFile("hdfs://laada-spark-master:54310/user/hduser/SparkClara/" + testDataPath, procNum);
		mData = data.map(new ParseData());
		JavaRDD<List<Point>> chunks = mData.groupBy(new DataGrouper(NUM_OF_EXPERIMENTS)).values();
		chunks.persist(StorageLevel.MEMORY_ONLY());
		List<List<Point>> medoids = chunks.map(new PamClara(DATA_SAMPLE_SIZE, NUM_OF_MEDOIDS)).collect();
		List<List<Double>> costs = chunks.map(new GetClusterCost(medoids)).collect();
		List<Point> bestSolution = getBestSolution(costs, medoids);
		sc.parallelize(bestSolution).saveAsTextFile("hdfs://laada-spark-master:54310/user/hduser/SparkClara/result/" + dateFormat.format(date));
		System.out.println("-----RESULT TIME: " + ((System.currentTimeMillis() - start) / 1000));
	}
	
	private static List<Point> getBestSolution(List<List<Double>> costs, List<List<Point>> medoids)
	{
		List<Double> dist = calcFinalDist(costs);
		return medoids.get(getMinDistIdx(dist));
	}
	
	private static int getMinDistIdx(List<Double> dist)
	{
		double min = Double.MAX_VALUE;
		int idx = 0;
		
		for (int i = 0; i < dist.size(); i++)
		{
			if(dist.get(i) < min) 
			{
				min = dist.get(i);
				idx = i;
			}
		}
		
		return idx;
	}
	
	private static List<Double> calcFinalDist(List<List<Double>> costs)
	{
		List<Double> result = new ArrayList<Double>();
		
		for (int i = 0; i < costs.size(); i++)
		{
			result.add(0.0);
			
			for (int j = 0; j < costs.get(i).size(); j++)
			{
				result.set(i, result.get(i) + costs.get(i).get(j));
			}
		}
		
		return result;
	}
}

@SuppressWarnings("serial")
class GetClusterCost extends Function<List<Point>, List<Double>> {
	
	private List<List<Point>> mMedoids;
	
	GetClusterCost(List<List<Point>> medoids) {
		mMedoids = medoids;
	}
	
	/**
	 * Process data
	 */
	public List<Double> call(List<Point> value) {	
		ArrayList<Double> result = new ArrayList<Double>();
		
		for (int i = 0; i < mMedoids.size(); i++)
		{
			result.add(pointsToMedoids(value, mMedoids.get(i)));
		}
		
		return result;
	}
	
	private double pointsToMedoids(List<Point> points, List<Point> medoids) {
		double result = 0;
		
		for (int i = 0; i < points.size(); i++)
		{
			result += getPointDist(points.get(i), medoids);
		}
		
		return result;
	}
	
	private double getPointDist(Point point, List<Point> medoids) {
		double min = Double.MAX_VALUE;
		Point medoid;
		double distance;
		
		for (int i = 0; i < medoids.size(); i++)
		{
			medoid = medoids.get(i);
			distance = calcDistance(point.getX(), medoid.getX(), point.getY(), medoid.getY());
			if(distance < min) min = distance;
		}
		
		return min;
	}
	
	private double calcDistance(long x1, long x2, long y1, long y2) {
		// Double.parseDouble(new
		// DecimalFormat("#.##").format(unformattedNumber));
		return Math.sqrt(Math.pow((x2 - x1), 2) + Math.pow((y2 - y1), 2));
	}
}

/**
 * Parse data class
 * @author SergeiL
 */
@SuppressWarnings("serial")
class ParseData extends Function<String, Point> {
	
	/**
	 * Process data
	 */
	public Point call(String value) {
		String[] valueArr = value.split(";");
		return new Point(Integer.parseInt(valueArr[0]), Integer.parseInt(valueArr[1]));
	}
}